@@ -28,19 +28,27 @@ import emscripten
 # Minimal subset of targets used by CI systems to build enough to useful
 MINIMAL_TASKS = [
     'libcompiler_rt',
+    'libcompiler_rt-wasm-sjlj',
     'libc',
     'libc-debug',
-    'libc-optz',
-    'libc-optz-debug',
+    'libc_optz',
+    'libc_optz-debug',
     'libc++abi',
     'libc++abi-except',
     'libc++abi-noexcept',
+    'libc++abi-debug',
+    'libc++abi-debug-except',
+    'libc++abi-debug-noexcept',
     'libc++',
     'libc++-except',
     'libc++-noexcept',
     'libal',
     'libdlmalloc',
+    'libdlmalloc-noerrno',
+    'libdlmalloc-tracing',
     'libdlmalloc-debug',
+    'libembind',
+    'libembind-rtti',
     'libemmalloc',
     'libemmalloc-debug',
     'libemmalloc-memvalidate',
@@ -54,25 +62,42 @@ MINIMAL_TASKS = [
     'struct_info',
     'libstandalonewasm',
     'crt1',
+    'crt1_proxy_main',
     'libunwind-except',
     'libnoexit',
+    'sqlite3',
+    'sqlite3-mt',
 ]
 
-# Variant builds that we want to support for certain ports
-# TODO: It would be nice if the ports themselves could specify the variants that they
-# support.
-PORT_VARIANTS = {
-    'regal-mt': ('regal', {'USE_PTHREADS': 1}),
-    'harfbuzz-mt': ('harfbuzz', {'USE_PTHREADS': 1}),
-    'sdl2-mt': ('sdl2', {'USE_PTHREADS': 1}),
-    'icu-mt': ('icu', {'USE_PTHREADS': 1}),
-    'sdl2_mixer_mp3': ('sdl2_mixer', {'SDL2_MIXER_FORMATS': ["mp3"]}),
-    'sdl2_mixer_none': ('sdl2_mixer', {'SDL2_MIXER_FORMATS': []}),
-    'sdl2_image_png': ('sdl2_image', {'SDL2_IMAGE_FORMATS': ["png"]}),
-    'sdl2_image_jpg': ('sdl2_image', {'SDL2_IMAGE_FORMATS': ["jpg"]}),
-}
+# Additional tasks on top of MINIMAL_TASKS that are necessary for PIC testing on
+# CI (which has slightly more tests than other modes that want to use MINIMAL)
+MINIMAL_PIC_TASKS = MINIMAL_TASKS + [
+    'libcompiler_rt-mt',
+    'libc-mt',
+    'libc-mt-debug',
+    'libc_optz-mt',
+    'libc_optz-mt-debug',
+    'libc++abi-mt',
+    'libc++abi-mt-noexcept',
+    'libc++abi-debug-mt',
+    'libc++abi-debug-mt-noexcept',
+    'libc++-mt',
+    'libc++-mt-noexcept',
+    'libdlmalloc-mt',
+    'libGL-emu',
+    'libGL-mt',
+    'libsockets_proxy',
+    'libsockets-mt',
+    'crtbegin',
+    'libsanitizer_common_rt',
+    'libubsan_rt',
+    'libwasm_workers_stub-debug',
+    'libwebgpu_cpp',
+    'libfetch',
+    'libwasmfs',
+]
 
-PORTS = sorted(list(ports.ports_by_name.keys()) + list(PORT_VARIANTS.keys()))
+PORTS = sorted(list(ports.ports_by_name.keys()) + list(ports.port_variants.keys()))
 
 temp_files = shared.get_temp_files()
 logger = logging.getLogger('embuilder')
@@ -95,8 +120,8 @@ Issuing 'embuilder build ALL' causes each task to be built.
 
 @contextmanager
 def get_port_variant(name):
-  if name in PORT_VARIANTS:
-    name, extra_settings = PORT_VARIANTS[name]
+  if name in ports.port_variants:
+    name, extra_settings = ports.port_variants[name]
     old_settings = settings.dict().copy()
     for key, value in extra_settings.items():
       setattr(settings, key, value)
@@ -191,6 +216,9 @@ def main():
   elif 'MINIMAL' in tasks:
     tasks = MINIMAL_TASKS
     auto_tasks = True
+  elif 'MINIMAL_PIC' in tasks:
+    tasks = MINIMAL_PIC_TASKS
+    auto_tasks = True
   elif 'ALL' in tasks:
     tasks = system_tasks + PORTS
     auto_tasks = True
@@ -214,7 +242,7 @@ def main():
       if do_clear:
         library.erase()
       if do_build:
-        library.get_path()
+        library.build(deterministic_paths=True)
     elif what == 'sysroot':
       if do_clear:
         shared.Cache.erase_file('sysroot_install.stamp')
