@@ -16,7 +16,7 @@ LibraryJSEventLoop = {
   $setImmediateWrapped: function(func) {
     if (!setImmediateWrapped.mapping) setImmediateWrapped.mapping = [];
     var id = setImmediateWrapped.mapping.length;
-    setImmediateWrapped.mapping[id] = setImmediate(function() {
+    setImmediateWrapped.mapping[id] = setImmediate(() => {
       setImmediateWrapped.mapping[id] = undefined;
       func();
     });
@@ -34,33 +34,35 @@ LibraryJSEventLoop = {
   },
 
   $polyfillSetImmediate__deps: ['$setImmediateWrapped', '$clearImmediateWrapped'],
-  $polyfillSetImmediate__postset:
-    'var emSetImmediate;\n' +
-    'var emClearImmediate;\n' +
-    'if (typeof setImmediate !== "undefined") {\n' +
-      'emSetImmediate = setImmediateWrapped;\n' +
-      'emClearImmediate = clearImmediateWrapped;\n' +
-    '} else if (typeof addEventListener === "function") {\n' +
-      'var __setImmediate_id_counter = 0;\n' +
-      'var __setImmediate_queue = [];\n' +
-      'var __setImmediate_message_id = "_si";\n' +
-      'function __setImmediate_cb(e) {\n' +
-        'if (e.data === __setImmediate_message_id) {\n' +
-          'e.stopPropagation();\n' +
-          '__setImmediate_queue.shift()();\n' +
-          '++__setImmediate_id_counter;\n' +
-        '}\n' +
-      '}\n' +
-      'addEventListener("message", __setImmediate_cb, true);\n' +
-      'emSetImmediate = function(func) {\n' +
-        'postMessage(__setImmediate_message_id, "*");\n' +
-        'return __setImmediate_id_counter + __setImmediate_queue.push(func) - 1;\n' +
-      '}\n' +
-      'emClearImmediate = /**@type{function(number=)}*/(function(id) {\n' +
-        'var index = id - __setImmediate_id_counter;\n' +
-        'if (index >= 0 && index < __setImmediate_queue.length) __setImmediate_queue[index] = function() {};\n' + // must preserve the order and count of elements in the queue, so replace the pending callback with an empty function
-      '})\n' +
-    '}',
+  $polyfillSetImmediate__postset: `
+    var emSetImmediate;
+    var emClearImmediate;
+    if (typeof setImmediate != "undefined") {
+      emSetImmediate = setImmediateWrapped;
+      emClearImmediate = clearImmediateWrapped;
+    } else if (typeof addEventListener == "function") {
+      var __setImmediate_id_counter = 0;
+      var __setImmediate_queue = [];
+      var __setImmediate_message_id = "_si";
+      /** @param {Event} e */
+      var __setImmediate_cb = (e) => {
+        if (e.data === __setImmediate_message_id) {
+          e.stopPropagation();
+          __setImmediate_queue.shift()();
+          ++__setImmediate_id_counter;
+        }
+      }
+      addEventListener("message", __setImmediate_cb, true);
+      emSetImmediate = (func) => {
+        postMessage(__setImmediate_message_id, "*");
+        return __setImmediate_id_counter + __setImmediate_queue.push(func) - 1;
+      }
+      emClearImmediate = /**@type{function(number=)}*/((id) => {
+        var index = id - __setImmediate_id_counter;
+        // must preserve the order and count of elements in the queue, so replace the pending callback with an empty function
+        if (index >= 0 && index < __setImmediate_queue.length) __setImmediate_queue[index] = () => {};
+      })
+    }`,
 
   $polyfillSetImmediate: function() {
     // nop, used for its postset to ensure setImmediate() polyfill is
@@ -68,11 +70,8 @@ LibraryJSEventLoop = {
     // emscripten_set_immediate_loop() if application links to both of them.
   },
 
-  emscripten_set_immediate__deps: ['$polyfillSetImmediate', '$callUserCallback',
-#if !MINIMAL_RUNTIME
-    '$runtimeKeepalivePush', '$runtimeKeepalivePop',
-#endif
-  ],
+  emscripten_set_immediate__sig: 'ipp',
+  emscripten_set_immediate__deps: ['$polyfillSetImmediate', '$callUserCallback'],
   emscripten_set_immediate: function(cb, userData) {
     polyfillSetImmediate();
     {{{ runtimeKeepalivePush(); }}}
@@ -84,21 +83,15 @@ LibraryJSEventLoop = {
     });
   },
 
-  emscripten_clear_immediate__deps: ['$polyfillSetImmediate',
-#if !MINIMAL_RUNTIME
-    '$runtimeKeepalivePop',
-#endif
-  ],
+  emscripten_clear_immediate__sig: 'vi',
+  emscripten_clear_immediate__deps: ['$polyfillSetImmediate'],
   emscripten_clear_immediate: function(id) {
     {{{ runtimeKeepalivePop(); }}}
     emClearImmediate(id);
   },
 
-  emscripten_set_immediate_loop__deps: ['$polyfillSetImmediate', '$callUserCallback',
-#if !MINIMAL_RUNTIME
-    '$runtimeKeepalivePush', '$runtimeKeepalivePop',
-#endif
-  ],
+  emscripten_set_immediate_loop__sig: 'vpp' ,
+  emscripten_set_immediate_loop__deps: ['$polyfillSetImmediate', '$callUserCallback'],
   emscripten_set_immediate_loop: function(cb, userData) {
     polyfillSetImmediate();
     function tick() {
@@ -114,30 +107,19 @@ LibraryJSEventLoop = {
     return emSetImmediate(tick);
   },
 
-  emscripten_set_timeout__deps: ['$callUserCallback',
-#if !MINIMAL_RUNTIME
-    '$runtimeKeepalivePush', '$runtimeKeepalivePop',
-#endif
-  ],
+  emscripten_set_timeout__sig: 'ipdp',
+  emscripten_set_timeout__deps: ['$safeSetTimeout'],
   emscripten_set_timeout: function(cb, msecs, userData) {
-    {{{ runtimeKeepalivePush() }}}
-    return setTimeout(function() {
-      {{{ runtimeKeepalivePop() }}}
-      callUserCallback(function() {
-        {{{ makeDynCall('vi', 'cb') }}}(userData);
-      });
-    }, msecs);
+    return safeSetTimeout(() => {{{ makeDynCall('vi', 'cb') }}}(userData), msecs);
   },
 
+  emscripten_clear_timeout__sig: 'vi',
   emscripten_clear_timeout: function(id) {
     clearTimeout(id);
   },
 
-  emscripten_set_timeout_loop__deps: ['$callUserCallback',
-#if !MINIMAL_RUNTIME
-    '$runtimeKeepalivePush', '$runtimeKeepalivePop',
-#endif
-  ],
+  emscripten_set_timeout_loop__sig: 'vpdp',
+  emscripten_set_timeout_loop__deps: ['$callUserCallback'],
   emscripten_set_timeout_loop: function(cb, msecs, userData) {
     function tick() {
       var t = performance.now();
@@ -157,11 +139,8 @@ LibraryJSEventLoop = {
     return setTimeout(tick, 0);
   },
 
-  emscripten_set_interval__deps: ['$callUserCallback',
-#if !MINIMAL_RUNTIME
-    '$runtimeKeepalivePush', '$runtimeKeepalivePop',
-#endif
-  ],
+  emscripten_set_interval__sig: 'ipdp',
+  emscripten_set_interval__deps: ['$callUserCallback'],
   emscripten_set_interval: function(cb, msecs, userData) {
     {{{ runtimeKeepalivePush() }}}
     return setInterval(function() {
@@ -171,9 +150,7 @@ LibraryJSEventLoop = {
     }, msecs);
   },
 
-#if !MINIMAL_RUNTIME
-  emscripten_clear_interval__deps: ['$runtimeKeepalivePop'],
-#endif
+  emscripten_clear_interval__sig: 'vi',
   emscripten_clear_interval: function(id) {
     {{{ runtimeKeepalivePop() }}}
     clearInterval(id);
