@@ -5,9 +5,14 @@
  */
 
 #if USE_CLOSURE_COMPILER
-// if (!Module)` is crucial for Closure Compiler here as it will otherwise replace every `Module` occurrence with the object below
+// if (!Module)` is crucial for Closure Compiler here as it will
+// otherwise replace every `Module` occurrence with the object below
 var /** @type{Object} */ Module;
 if (!Module) /** @suppress{checkTypes}*/Module = {"__EMSCRIPTEN_PRIVATE_MODULE_EXPORT_NAME_SUBSTITUTION__":1};
+#elif ENVIRONMENT_MAY_BE_NODE || ENVIRONMENT_MAY_BE_SHELL
+// When running on the web we expect Module to be defined externally, in the
+// HTML.  Otherwise we must define it here before its first use
+var Module = typeof {{{ EXPORT_NAME }}} != 'undefined' ? {{{ EXPORT_NAME }}} : {};
 #else
 var Module = {{{ EXPORT_NAME }}};
 #endif // USE_CLOSURE_COMPILER
@@ -99,13 +104,8 @@ if (ENVIRONMENT_IS_SHELL) {
 // and build with --closure 1 to get Closure optimize out all the uses
 // altogether.
 
-function out(text) {
-  console.log(text);
-}
-
-function err(text) {
-  console.error(text);
-}
+var out = (text) => console.log(text);
+var err = (text) => console.error(text);
 
 // Override this function in a --pre-js file to get a signal for when
 // compilation is ready. In that callback, call the function run() to start
@@ -115,9 +115,9 @@ function ready() {
   readyPromiseResolve(Module);
 #endif // MODULARIZE
 #if INVOKE_RUN && HAS_MAIN
-  {{{ runOnMainThread("run();") }}}
+  {{{ runIfMainThread("run();") }}}
 #elif ASSERTIONS
-  console.log('ready() called, and INVOKE_RUN=0. The runtime is now ready for you to call run() to invoke application _main(). You can also override ready() in a --pre-js file to get this signal as a callback')
+  out('ready() called, and INVOKE_RUN=0. The runtime is now ready for you to call run() to invoke application _main(). You can also override ready() in a --pre-js file to get this signal as a callback')
 #endif
 }
 
