@@ -22,9 +22,7 @@ if (ENVIRONMENT_IS_NODE) {
 
   var parentPort = nodeWorkerThreads.parentPort;
 
-  parentPort.on('message', function(data) {
-    onmessage({ data: data });
-  });
+  parentPort.on('message', (data) => onmessage({ data: data }));
 
   var fs = require('fs');
 
@@ -51,10 +49,12 @@ if (ENVIRONMENT_IS_NODE) {
 }
 #endif // ENVIRONMENT_MAY_BE_NODE
 
-// Thread-local:
-#if EMBIND
-var initializedJS = false; // Guard variable for one-time init of the JS state (currently only embind types registration)
-#endif
+// Thread-local guard variable for one-time init of the JS state
+var initializedJS = false;
+
+// Proxying queues that were notified before the thread started and need to be
+// executed as part of startup.
+var pendingNotifiedProxyingQueues = [];
 
 #if ASSERTIONS
 function assert(condition, text) {
@@ -85,6 +85,9 @@ var out = () => { throw 'out() is not defined in worker.js.'; }
 #endif
 var err = threadPrintErr;
 self.alert = threadAlert;
+#if RUNTIME_DEBUG
+var dbg = threadPrintErr;
+#endif
 
 #if !MINIMAL_RUNTIME
 Module['instantiateWasm'] = (info, receiveInstance) => {
@@ -105,11 +108,17 @@ Module['instantiateWasm'] = (info, receiveInstance) => {
 }
 #endif
 
+// Turn unhandled rejected promises into errors so that the main thread will be
+// notified about them.
+self.onunhandledrejection = (e) => {
+  throw e.reason ?? e;
+};
+
 self.onmessage = (e) => {
   try {
     if (e.data.cmd === 'load') { // Preload command that is called once per worker to parse and load the Emscripten code.
 #if PTHREADS_DEBUG
-      err('worker.js: loading module')
+      dbg('worker.js: loading module')
 #endif
 #if MINIMAL_RUNTIME
       var imports = {};
@@ -130,6 +139,14 @@ self.onmessage = (e) => {
       Module['dynamicLibraries'] = e.data.dynamicLibraries;
 #endif
 
+      // Use `const` here to ensure that the variable is scoped only to
+      // that iteration, allowing safe reference from a closure.
+      for (const handler of e.data.handlers) {
+        Module[handler] = function() {
+          postMessage({ cmd: 'callHandler', handler, args: [...arguments] });
+        }
+      }
+
       {{{ makeAsmImportsAccessInPthread('wasmMemory') }}} = e.data.wasmMemory;
 
 #if LOAD_SOURCE_MAP
@@ -200,54 +217,40 @@ self.onmessage = (e) => {
       // (+/- 0.1msecs in testing).
       Module['__performance_now_clock_drift'] = performance.now() - e.data.time;
 
-      // Pass the thread address inside the asm.js scope to store it for fast access that avoids the need for a FFI out.
-      Module['__emscripten_thread_init'](e.data.threadInfoStruct, /*isMainBrowserThread=*/0, /*isMainRuntimeThread=*/0, /*canBlock=*/1);
+      // Pass the thread address to wasm to store it for fast access.
+      Module['__emscripten_thread_init'](e.data.pthread_ptr, /*isMainBrowserThread=*/0, /*isMainRuntimeThread=*/0, /*canBlock=*/1);
 
 #if ASSERTIONS
-      assert(e.data.threadInfoStruct);
+      assert(e.data.pthread_ptr);
 #endif
       // Also call inside JS module to set up the stack frame for this pthread in JS module scope
       Module['establishStackSpace']();
       Module['PThread'].receiveObjectTransfer(e.data);
-      Module['PThread'].threadInit();
+      Module['PThread'].threadInitTLS();
 
-#if EMBIND
-      // Embind must initialize itself on all threads, as it generates support JS.
-      // We only do this once per worker since they get reused
       if (!initializedJS) {
-        Module['___embind_register_native_and_builtin_types']();
-        initializedJS = true;
-      }
+#if EMBIND
+#if PTHREADS_DEBUG
+        dbg('Pthread 0x' + Module['_pthread_self']().toString(16) + ' initializing embind.');
+#endif
+        // Embind must initialize itself on all threads, as it generates support JS.
+        // We only do this once per worker since they get reused
+        Module['__embind_initialize_bindings']();
 #endif // EMBIND
 
-      try {
-        // pthread entry points are always of signature 'void *ThreadMain(void *arg)'
-        // Native codebases sometimes spawn threads with other thread entry point signatures,
-        // such as void ThreadMain(void *arg), void *ThreadMain(), or void ThreadMain().
-        // That is not acceptable per C/C++ specification, but x86 compiler ABI extensions
-        // enable that to work. If you find the following line to crash, either change the signature
-        // to "proper" void *ThreadMain(void *arg) form, or try linking with the Emscripten linker
-        // flag -sEMULATE_FUNCTION_POINTER_CASTS to add in emulation for this x86 ABI extension.
-        var result = Module['invokeEntryPoint'](e.data.start_routine, e.data.arg);
+        // Execute any proxied work that came in before the thread was
+        // initialized. Only do this once because it is only possible for
+        // proxying notifications to arrive before thread initialization on
+        // fresh workers.
+        pendingNotifiedProxyingQueues.forEach(queue => {
+          Module['executeNotifiedProxyingQueue'](queue);
+        });
+        pendingNotifiedProxyingQueues = [];
+        initializedJS = true;
+      }
 
-#if STACK_OVERFLOW_CHECK
-        Module['checkStackCookie']();
-#endif
-#if MINIMAL_RUNTIME
-        // In MINIMAL_RUNTIME the noExitRuntime concept does not apply to
-        // pthreads. To exit a pthread with live runtime, use the function
-        // emscripten_unwind_to_js_event_loop() in the pthread body.
-        // The thread might have finished without calling pthread_exit(). If so,
-        // then perform the exit operation ourselves.
-        // (This is a no-op if explicit pthread_exit() had been called prior.)
-        Module['__emscripten_thread_exit'](result);
-#else
-        if (Module['keepRuntimeAlive']()) {
-          Module['PThread'].setExitStatus(result);
-        } else {
-          Module['__emscripten_thread_exit'](result);
-        }
-#endif
+      try {
+        Module['invokeEntryPoint'](e.data.start_routine, e.data.arg);
       } catch(ex) {
         if (ex != 'unwind') {
           // ExitStatus not present in MINIMAL_RUNTIME
@@ -281,19 +284,29 @@ self.onmessage = (e) => {
       }
     } else if (e.data.cmd === 'cancel') { // Main thread is asking for a pthread_cancel() on this thread.
       if (Module['_pthread_self']()) {
-        Module['__emscripten_thread_exit'](-1/*PTHREAD_CANCELED*/);
+        Module['__emscripten_thread_exit']({{{ cDefine('PTHREAD_CANCELED') }}});
       }
     } else if (e.data.target === 'setimmediate') {
       // no-op
     } else if (e.data.cmd === 'processProxyingQueue') {
-      executeNotifiedProxyingQueue(e.data.queue);
-    } else {
+      if (initializedJS) {
+        Module['executeNotifiedProxyingQueue'](e.data.queue);
+      } else {
+        // Defer executing this queue until the runtime is initialized.
+        pendingNotifiedProxyingQueues.push(e.data.queue);
+      }
+    } else if (e.data.cmd) {
+      // The received message looks like something that should be handled by this message
+      // handler, (since there is a e.data.cmd field present), but is not one of the
+      // recognized commands:
       err('worker.js received unknown command ' + e.data.cmd);
       err(e.data);
     }
   } catch(ex) {
+#if ASSERTIONS
     err('worker.js onmessage() captured an uncaught exception: ' + ex);
     if (ex && ex.stack) err(ex.stack);
+#endif
     if (Module['__emscripten_thread_crashed']) {
       Module['__emscripten_thread_crashed']();
     }
