@@ -246,7 +246,7 @@ def inspect_headers(headers, cflags):
   # TODO(sbc): If we can remove EM_EXCLUSIVE_CACHE_ACCESS then this would not longer be needed.
   shared.check_sanity()
 
-  compiler_rt = system_libs.Library.get_usable_variations()['libcompiler_rt'].get_path()
+  compiler_rt = system_libs.Library.get_usable_variations()['libcompiler_rt'].build()
 
   # Close all unneeded FDs.
   os.close(src_file[0])
@@ -261,17 +261,17 @@ def inspect_headers(headers, cflags):
   else:
     compiler = shared.EMCC
 
-  # -Oz optimizes enough to avoid warnings on code size/num locals
+  # -O1+ produces calls to iprintf, which libcompiler_rt doesn't support
   cmd = [compiler] + cflags + ['-o', js_file[1], src_file[1],
                                '-O0',
                                '-Werror',
                                '-Wno-format',
                                '-nostdlib',
                                compiler_rt,
-                               '-sBOOTSTRAPPING_STRUCT_INFO=1',
-                               '-sLLD_REPORT_UNDEFINED=1',
+                               '-sBOOTSTRAPPING_STRUCT_INFO',
+                               '-sLLD_REPORT_UNDEFINED',
                                '-sSTRICT',
-                               # Use SINGLE_FILE=1 so there is only a single
+                               # Use SINGLE_FILE so there is only a single
                                # file to cleanup.
                                '-sSINGLE_FILE']
 
@@ -288,7 +288,7 @@ def inspect_headers(headers, cflags):
   if settings.MEMORY64:
     # Always use =2 here so that we don't generate binar that actually requires
     # memeory64 to run.  All we care about is that the output is correct.
-    cmd += ['-sMEMORY64=2']
+    cmd += ['-sMEMORY64=2', '-Wno-experimental']
 
   show(shared.shlex_join(cmd))
   try:
@@ -301,7 +301,7 @@ def inspect_headers(headers, cflags):
   show('Calling generated program... ' + js_file[1])
   args = []
   if settings.MEMORY64:
-    args += ['--experimental-wasm-bigint']
+    args += shared.node_bigint_flags()
   info = shared.run_js_tool(js_file[1], node_args=args, stdout=shared.PIPE).splitlines()
 
   if not DEBUG:
@@ -434,12 +434,13 @@ def main(args):
     '-I' + utils.path_from_root('system/lib/libc/musl/src/internal'),
     '-I' + utils.path_from_root('system/lib/libc/musl/src/include'),
     '-I' + utils.path_from_root('system/lib/pthread/'),
-    '-I' + utils.path_from_root('system/lib/wasmfs/'),
   ]
 
   cxxflags = [
     '-I' + utils.path_from_root('system/lib/libcxxabi/src'),
     '-D__USING_EMSCRIPTEN_EXCEPTIONS__',
+    '-I' + utils.path_from_root('system/lib/wasmfs/'),
+    '-std=c++17',
   ]
 
   # Look for structs in all passed headers.
