@@ -4,92 +4,72 @@
 # found in the LICENSE file.
 
 import os
-import shutil
 import logging
 import re
 
 TAG = 'version_3_3'
 HASH = 'd7b22660036c684f09754fcbbc7562984f02aa955eef2b76555270c63a717e6672c4fe695afb16280822e8b7c75d4b99ae21975a01a4ed51cad957f7783722cd'
 
+deps = ['libpng', 'zlib']
+
+
+def needed(settings):
+  return settings.USE_COCOS2D == 3
+
 
 def get(ports, settings, shared):
-  if settings.USE_COCOS2D != 3:
-    return []
+  ports.fetch_project('cocos2d', f'https://github.com/emscripten-ports/Cocos2d/archive/{TAG}.zip', sha512hash=HASH)
 
-  ports.fetch_project(
-    'cocos2d', 'https://github.com/emscripten-ports/Cocos2d/archive/' + TAG + '.zip', 'Cocos2d-' + TAG, sha512hash=HASH)
-  libname = ports.get_lib_name('libcocos2d')
-
-  def create():
-    logging.info('building port: cocos2d v3')
+  def create(final):
     logging.warn('cocos2d: library is experimental, do not expect that it will work out of the box')
 
     cocos2d_src = os.path.join(ports.get_dir(), 'cocos2d')
     cocos2d_root = os.path.join(cocos2d_src, 'Cocos2d-' + TAG)
     cocos2dx_root = os.path.join(cocos2d_root, 'cocos2dx')
-    cocos2dx_src = make_source_list(cocos2d_root, cocos2dx_root)
-    cocos2dx_includes = make_includes(cocos2d_root)
 
-    cocos2d_build = os.path.join(ports.get_build_dir(), 'cocos2d')
-    shared.try_delete(os.path.join(cocos2d_build, 'samples'))
-    shutil.copytree(os.path.join(cocos2d_root, 'samples', 'Cpp'),
-                    os.path.join(cocos2d_build, 'samples'))
+    srcs = make_source_list(cocos2d_root, cocos2dx_root)
+    includes = make_includes(cocos2d_root)
+    flags = [
+      '-w',
+      '-D__CC_PLATFORM_FILEUTILS_CPP__',
+      '-DCC_ENABLE_CHIPMUNK_INTEGRATION',
+      '-DCC_KEYBOARD_SUPPORT',
+      '-DGL_ES=1',
+      '-DNDEBUG', # '-DCOCOS2D_DEBUG=1' 1 - error/warn, 2 - verbose
+      # Cocos2d source code hasn't switched to __EMSCRIPTEN__.
+      # See https://github.com/emscripten-ports/Cocos2d/pull/3
+      '-DEMSCRIPTEN',
+      '-DCP_USE_DOUBLES=0',
+      '-sUSE_ZLIB',
+      '-sUSE_LIBPNG',
+    ]
 
-    commands = []
-    o_s = []
-    for src in cocos2dx_src:
-      o = os.path.join(cocos2d_build, 'Cocos2d-' + TAG, 'build', src + '.o')
-      shared.safe_ensure_dirs(os.path.dirname(o))
-      command = [shared.PYTHON,
-                 shared.EMCC,
-                 '-c', src,
-                 '-Wno-overloaded-virtual',
-                 '-Wno-deprecated-declarations',
-                 '-D__CC_PLATFORM_FILEUTILS_CPP__',
-                 '-DCC_ENABLE_CHIPMUNK_INTEGRATION',
-                 '-DCC_KEYBOARD_SUPPORT',
-                 '-DGL_ES=1',
-                 '-DNDEBUG', # '-DCOCOS2D_DEBUG=1' 1 - error/warn, 2 - verbose
-                 '-DCP_USE_DOUBLES=0',
-                 '-O2',
-                 '-s', 'USE_ZLIB=1',
-                 '-s', 'USE_LIBPNG=1',
-                 '-o', o, '-w']
-
-      for include in cocos2dx_includes:
-        command.append('-I' + include)
-
-      commands.append(command)
-      o_s.append(o)
-    shared.safe_ensure_dirs(os.path.dirname(o_s[0]))
-    ports.run_commands(commands)
-    final = os.path.join(cocos2d_build, libname)
-    ports.create_lib(final, o_s)
-
-    for dirname in cocos2dx_includes:
+    for dirname in includes:
       target = os.path.join('cocos2d', os.path.relpath(dirname, cocos2d_root))
       ports.install_header_dir(dirname, target=target)
 
-    return final
+    ports.build_port(cocos2d_src, final, 'cocos2d',
+                     flags=flags,
+                     cxxflags=['-std=c++14'],
+                     includes=includes,
+                     srcs=srcs)
 
-  return [shared.Cache.get(libname, create, what='port')]
+  return [shared.Cache.get_lib('libcocos2d.a', create, what='port')]
 
 
-def clear(ports, shared):
-  shared.Cache.erase_file(ports.get_lib_name('libcocos2d'))
+def clear(ports, settings, shared):
+  shared.Cache.erase_lib('libcocos2d.a')
 
 
 def process_dependencies(settings):
-  if settings.USE_COCOS2D == 3:
-    settings.USE_LIBPNG = 1
-    settings.USE_ZLIB = 1
+  settings.USE_LIBPNG = 1
+  settings.USE_ZLIB = 1
 
 
-def process_args(ports, args, settings, shared):
-  if settings.USE_COCOS2D == 3:
-    get(ports, settings, shared)
-    for include in make_includes(os.path.join(ports.get_include_dir(), 'cocos2d')):
-      args.append('-I' + include)
+def process_args(ports):
+  args = []
+  for include in make_includes(ports.get_include_dir('cocos2d')):
+    args.append('-I' + include)
   return args
 
 
