@@ -4,10 +4,9 @@
 # found in the LICENSE file.
 
 import os
-import shutil
 
-TAG = 'version_1'
-HASH = '0d0b1280ba0501ad0a23cf1daa1f86821c722218b59432734d3087a89acd22aabd5c3e5e1269700dcd41e87073046e906060f167c032eb91a3ac8c5808a02783'
+TAG = 'VER-2-6'
+HASH = '24c1706045209a3fac64f423116318f3f97892b984e89f05600c71ac356d48d78d642d18b739f26d51d7b7723821becb991a90175accfe7a91ca16a3c759e4f8'
 
 
 def needed(settings):
@@ -15,18 +14,13 @@ def needed(settings):
 
 
 def get(ports, settings, shared):
-  ports.fetch_project('freetype', 'https://github.com/emscripten-ports/FreeType/archive/' + TAG + '.zip', 'FreeType-' + TAG, sha512hash=HASH)
+  ports.fetch_project('freetype', f'https://github.com/freetype/freetype/archive/{TAG}.zip', sha512hash=HASH)
 
   def create(final):
-    ports.clear_project_build('freetype')
-
-    source_path = os.path.join(ports.get_dir(), 'freetype', 'FreeType-' + TAG)
-    dest_path = os.path.join(ports.get_build_dir(), 'freetype')
-    shared.try_delete(dest_path)
-    os.makedirs(dest_path)
-    shutil.rmtree(dest_path, ignore_errors=True)
-    shutil.copytree(source_path, dest_path)
-    open(os.path.join(dest_path, 'include/ftconfig.h'), 'w').write(ftconf_h)
+    source_path = os.path.join(ports.get_dir(), 'freetype', 'freetype-' + TAG)
+    ports.write_file(os.path.join(source_path, 'include/ftconfig.h'), ftconf_h)
+    ports.install_header_dir(os.path.join(source_path, 'include'),
+                             target=os.path.join('freetype2'))
 
     # build
     srcs = ['src/autofit/autofit.c',
@@ -81,31 +75,22 @@ def get(ports, settings, shared):
             'src/type1/type1.c',
             'src/type42/type42.c',
             'src/winfonts/winfnt.c']
-    commands = []
-    o_s = []
-    for src in srcs:
-      o = os.path.join(ports.get_build_dir(), 'freetype', src + '.o')
-      shared.safe_ensure_dirs(os.path.dirname(o))
-      commands.append([shared.EMCC, '-c', os.path.join(dest_path, src), '-o', o,
-                       '-DFT2_BUILD_LIBRARY', '-O2',
-                       '-I' + dest_path + '/include',
-                       '-I' + dest_path + '/truetype',
-                       '-I' + dest_path + '/sfnt',
-                       '-I' + dest_path + '/autofit',
-                       '-I' + dest_path + '/smooth',
-                       '-I' + dest_path + '/raster',
-                       '-I' + dest_path + '/psaux',
-                       '-I' + dest_path + '/psnames',
-                       '-I' + dest_path + '/truetype',
-                       '-w'])
-      o_s.append(o)
 
-    ports.run_commands(commands)
-    shared.try_delete(final)
-    shared.run_process([shared.LLVM_AR, 'rc', final] + o_s)
+    flags = [
+      '-DFT2_BUILD_LIBRARY',
+      '-I' + source_path + '/include',
+      '-I' + source_path + '/truetype',
+      '-I' + source_path + '/sfnt',
+      '-I' + source_path + '/autofit',
+      '-I' + source_path + '/smooth',
+      '-I' + source_path + '/raster',
+      '-I' + source_path + '/psaux',
+      '-I' + source_path + '/psnames',
+      '-I' + source_path + '/truetype',
+      '-pthread'
+    ]
 
-    ports.install_header_dir(os.path.join(dest_path, 'include'),
-                             target=os.path.join('freetype2', 'freetype'))
+    ports.build_port(source_path, final, 'freetype', flags=flags, srcs=srcs)
 
   return [shared.Cache.get_lib('libfreetype.a', create, what='port')]
 
@@ -115,7 +100,7 @@ def clear(ports, settings, shared):
 
 
 def process_args(ports):
-  return ['-I' + os.path.join(ports.get_include_dir(), 'freetype2', 'freetype')]
+  return ['-I' + ports.get_include_dir('freetype2')]
 
 
 def show():
